/**
\author		Korotkov Andrey aka DRON
\date		14.03.2011 (c)Korotkov Andrey

This file is a part of DGLE2 project and is distributed
under the terms of the GNU Lesser General Public License.
See "DGLE2.h" for more details.
*/

#include "MusicMCI.h"

#pragma message("Linking with \"Winmm.lib\".")
#pragma comment(linker, "/defaultlib:Winmm.lib")

CMusicMCI::CMusicMCI(uint uiInstIdx, MCIDEVICEID uiDevId):
CInstancedObj(uiInstIdx), _uiVolume(100), _bIsPlaying(false),
_bIsLoop(false), _uiDevId(uiDevId)
{
	Core()->GetHandle(_stWnd);
	Core()->pDMessageProc()->Add(&_s_MessageProc, (void*)this);
}

CMusicMCI::CMusicMCI(uint uiInstIdx, IFile *pFile):
CInstancedObj(uiInstIdx), _uiVolume(100), _bIsPlaying(false),
_bIsLoop(false), _uiDevId(NULL)
{
	Core()->GetHandle(_stWnd);

	char name[64], path[256];

	if(FAILED(pFile->GetName(name, 64)) || FAILED(pFile->GetPath(path, 256)))
	{
		LOG("Can't get Name and Path of IFile.", LT_ERROR);
		throw((HRESULT)E_ABORT);
	}

	string file_name = string(path) + (path[strlen(path) - 1] == '|' ? " " : "\\") + string(name);

	if(file_name.find("|") != string::npos)//unpack file
	{
		char tmp_path[MAX_PATH];
		
		if(0 == GetTempPath(MAX_PATH, tmp_path))
		{
			LOG("Failed to get temporary directory.", LT_ERROR);
			throw((HRESULT)E_ABORT);
		}		

		_tmpFileName = string(tmp_path) + string(name);
	
		IFileSystem *p_hddfs;
		
		Core()->pMainFS()->GetVirtualFileSystem(NULL, p_hddfs);
		
		IFile *p_tmp_file;
		
		if(S_OK != p_hddfs->OpenFile(_tmpFileName.c_str(), (E_FILE_SYSTEM_OPEN_FLAGS)(FSOF_WRITE | FSOF_TRUNK | FSOF_BINARY), p_tmp_file))
		{
			LOG("Failed to create temporary music file.", LT_ERROR);
			throw((HRESULT)E_ABORT);
		}		

		uint size, rw;

		pFile->GetSize(size);
		
		DGLE2::byte *p_buff = new DGLE2::byte[size];

		pFile->Read(p_buff, size, rw);

		if(rw != size)
		{
			LOG("Failed to read music file.", LT_ERROR);
			delete[] p_buff;
			throw((HRESULT)E_ABORT);
		}

		p_tmp_file->Write(p_buff, size, rw);
		
		delete[] p_buff;

		if(rw != size)
		{
			LOG("Failed to write temporary music file.", LT_ERROR);
			throw((HRESULT)E_ABORT);
		}

		p_tmp_file->Free();

		file_name = _tmpFileName;
	}

	MCI_OPEN_PARMSA st_open_parm;

	st_open_parm.dwCallback			= (DWORD)_stWnd;
	st_open_parm.lpstrDeviceType	=  NULL;
	st_open_parm.lpstrElementName	= (LPCSTR)file_name.c_str();
	st_open_parm.lpstrAlias			= (LPCSTR)"";
	st_open_parm.lpstrDeviceType	= (LPCSTR)"";
	st_open_parm.wDeviceID			=  NULL;

	MCIERROR err = mciSendCommandA(0, MCI_OPEN, MCI_OPEN_ELEMENT,(DWORD)&st_open_parm);

	if(!err)
		_uiDevId = st_open_parm.wDeviceID;
	else
	{
		_uiDevId = NULL;
		LOG("Error while openning MCI device.",LT_ERROR);
		throw((HRESULT)E_ABORT);
	}

	Core()->pDMessageProc()->Add(&_s_MessageProc, (void*)this);
}

CMusicMCI::~CMusicMCI()
{
	if(_uiDevId)
	{
		MCI_GENERIC_PARMS st_generic_params;
		st_generic_params.dwCallback=(DWORD)_stWnd;

		MCIERROR err = mciSendCommand(_uiDevId,MCI_CLOSE,MCI_WAIT,(DWORD)&st_generic_params);

		_uiDevId = NULL;

		if(err)
			LOG("Can't close MCI device for unknown reason.",LT_ERROR);
	}

	if(_tmpFileName.length() > 0)
		if(0 == DeleteFile(_tmpFileName.c_str()))
			LOG("Failed to delete temporary music file.", LT_WARNING);

	Core()->pDMessageProc()->Remove(&_s_MessageProc, (void*)this);
}

void CMusicMCI::_MessageProc(const TWinMessage &stMsg)
{
	if(_bIsPlaying && stMsg.message == MM_MCINOTIFY && stMsg.wParam != MCI_NOTIFY_ABORTED && stMsg.lParam==_uiDevId)
	{
		if(!_bIsLoop)
			_bIsPlaying = false;
		else
		{
			MCI_SEEK_PARMS st_seek_parms;
			MCIERROR err = mciSendCommand(_uiDevId, MCI_SEEK, MCI_SEEK_TO_START | MCI_WAIT, (DWORD)&st_seek_parms);
			if(err) return;			
			MCI_PLAY_PARMS st_play_parm;
			st_play_parm.dwCallback = (DWORD)_stWnd;
			err = mciSendCommand(_uiDevId, MCI_PLAY, MCI_NOTIFY, (DWORD)&st_play_parm);
		}
	}
}

void CALLBACK CMusicMCI::_s_MessageProc(void *pParametr, const TWinMessage &stMsg)
{
	PTHIS(CMusicMCI)->_MessageProc(stMsg);
}

HRESULT CALLBACK CMusicMCI::Play(bool bLooped)
{
	if(!_uiDevId)
		return S_FALSE;

	MCI_PLAY_PARMS st_play_parm;
	st_play_parm.dwCallback =(DWORD)_stWnd;
	st_play_parm.dwFrom = 0;
	
	MCIERROR err = mciSendCommand(_uiDevId, MCI_PLAY, MCI_FROM | MCI_NOTIFY, (DWORD)&st_play_parm);
	
	if(err)
	{
		LOG("Can't start MCI playback.", LT_ERROR);
		return S_FALSE;
	}

	_bIsPlaying = true;
	_bIsLoop	= bLooped;

	return S_OK;
}

HRESULT CALLBACK CMusicMCI::Pause(bool bPaused)
{
	if(!_uiDevId)
		return S_FALSE;

	MCIERROR err;

	if(bPaused)
	{
		MCI_GENERIC_PARMS st_generic_params;
		st_generic_params.dwCallback=(DWORD)_stWnd;
		err = mciSendCommand(_uiDevId, MCI_PAUSE, MCI_NOTIFY, (DWORD)&st_generic_params);
	}
	else
	{
		MCI_PLAY_PARMS st_play_parm;
		st_play_parm.dwCallback =(DWORD)_stWnd;
		err = mciSendCommand(_uiDevId, MCI_PLAY, MCI_NOTIFY, (DWORD)&st_play_parm);
	}

	if(err)
		return E_FAIL;
	else
		_bIsPlaying	= !bPaused;

	return S_OK;
}

HRESULT CALLBACK CMusicMCI::Stop()
{
	if(!_uiDevId)
		return S_FALSE;

	MCI_GENERIC_PARMS st_generic_params;
	st_generic_params.dwCallback=(DWORD)_stWnd;

	MCIERROR err = mciSendCommand(_uiDevId, MCI_STOP, MCI_NOTIFY, (DWORD)&st_generic_params);

	_bIsPlaying = false;

	if(err)
		return E_FAIL;

	return S_OK;
}

HRESULT CALLBACK CMusicMCI::IsPlaying(bool &bPlaying)
{
	bPlaying = _bIsPlaying;	
	return S_OK;
}

HRESULT CALLBACK CMusicMCI::SetVolume(uint uiVolume)
{
	if(!_uiDevId)
		return S_FALSE;

	if(uiVolume > 100)
		uiVolume = 100;

	MCI_DGV_SETAUDIO_PARMS st_dvg_parms;

	st_dvg_parms.dwCallback		= (DWORD)_stWnd;
	st_dvg_parms.dwItem			= MCI_DGV_SETAUDIO_VOLUME;
	st_dvg_parms.dwValue		= uiVolume*10;
	st_dvg_parms.dwOver			= 0;
	st_dvg_parms.lpstrAlgorithm	= NULL;
	st_dvg_parms.lpstrQuality	= NULL;

	MCIERROR err = mciSendCommand(_uiDevId, MCI_SETAUDIO, MCI_NOTIFY | MCI_DGV_SETAUDIO_VALUE | MCI_DGV_SETAUDIO_ITEM, (DWORD)&st_dvg_parms);

	if(err)
	{
		if(err == 274)//Command unsupported by device. As known, only MP3 is supported.
			return E_ABORT;
		else
			return E_FAIL;
	}
	else
	{
		_uiVolume = uiVolume;
		return S_OK;
	}
}

HRESULT CALLBACK CMusicMCI::GetVolume(uint &uiVolume)
{
	uiVolume = _uiVolume;
	return S_OK;
}

HRESULT CALLBACK CMusicMCI::SetCurrentPosition(uint uiPos)
{
	if(!_uiDevId)
		return S_FALSE;

	MCI_SEEK_PARMS st_seek_parms;
	
	st_seek_parms.dwTo = uiPos;

	MCIERROR err = mciSendCommand(_uiDevId, MCI_SEEK, MCI_TO | MCI_WAIT, (DWORD)&st_seek_parms);

	if(err)
		return E_FAIL;
	
	if(_bIsPlaying)
	{
		MCI_PLAY_PARMS st_play_parm;
		st_play_parm.dwCallback = (DWORD)_stWnd;
		err = mciSendCommand(_uiDevId, MCI_PLAY, MCI_NOTIFY, (DWORD)&st_play_parm);

		if(err)
			return S_FALSE;
	}

	return S_OK;
}

HRESULT CALLBACK CMusicMCI::GetCurrentPosition(uint &uiPos)
{
	uiPos = 0;

	if(!_uiDevId)
		return S_FALSE;

	MCI_STATUS_PARMS st_status_params;

	st_status_params.dwItem = MCI_STATUS_POSITION;

	MCIERROR err = mciSendCommand(_uiDevId, MCI_STATUS, MCI_WAIT | MCI_STATUS_ITEM, (DWORD)&st_status_params);

	if(err)
		return E_FAIL;

	uiPos = st_status_params.dwReturn;

	return S_OK;
}

HRESULT CALLBACK CMusicMCI::GetLength(uint &uiLength)
{
	uiLength = 0;

	if(!_uiDevId)
		return S_FALSE;

	MCI_STATUS_PARMS st_status_params;

	st_status_params.dwItem = MCI_STATUS_LENGTH;

	MCIERROR err = mciSendCommand(_uiDevId, MCI_STATUS, MCI_WAIT | MCI_STATUS_ITEM, (DWORD)&st_status_params);

	if(err)
		return E_FAIL;

	uiLength = st_status_params.dwReturn;

	return S_OK;
}

HRESULT CALLBACK CMusicMCI::Free()
{
	bool can_delete;

	Core()->pResMan()->RemoveResource(this, can_delete);

	if(can_delete)
	{
		delete this;
		return S_OK;
	}
	else
		return S_FALSE;
}

HRESULT CALLBACK CMusicMCI::GetType(E_ENG_OBJ_TYPE &eObjType)
{
	eObjType = EOT_MUSIC;
	return S_OK;
}