/**
\author		Korotkov Andrey aka DRON
\date		15.03.2011 (c)Korotkov Andrey

This file is a part of DGLE2 project and is distributed
under the terms of the GNU Lesser General Public License.
See "DGLE2.h" for more details.
*/

#include "SoundMCI.h"

struct TMMFile
{
private:
	uint _uiRefCounter;
	
public:
	string stFileName;
	DGLE2::byte *pData;
	uint uiFileSize;

	TMMFile():pData(NULL), uiFileSize(0), _uiRefCounter(1) {}

	TMMFile(const TMMFile &mmfile):
	stFileName(mmfile.stFileName),
	pData(mmfile.pData), uiFileSize(mmfile.uiFileSize)
	{
		const_cast<TMMFile *>(&mmfile)->_uiRefCounter++;
		_uiRefCounter = 1;
	}

	TMMFile(IFile *pFile):
	_uiRefCounter(1)
	{
		char name[64];
		pFile->GetName(name, 64);
		stFileName = string(name);
		pFile->GetSize(uiFileSize);
		pData = new DGLE2::byte[uiFileSize];
		uint tmp;
		pFile->Read(pData, uiFileSize, tmp);

		if(tmp != uiFileSize)
		{
			uiFileSize = 0;
			delete[] pData;
			pData = NULL;
		}
	}

	~TMMFile()
	{
		if(pData && uiFileSize > 0 && --_uiRefCounter == 0)
			delete[] pData;
	}

	TMMFile &operator=(const TMMFile &mmfile)
	{
		TMMFile tmp;
		tmp.stFileName = mmfile.stFileName;
		tmp.pData = mmfile.pData;
		tmp.uiFileSize = mmfile.uiFileSize;

		const_cast<TMMFile *>(&mmfile)->_uiRefCounter++;
		tmp._uiRefCounter = 1;

		return tmp;
	}
};

vector<TMMFile> MMFiles;
uint CurrentMMFileId = -1;

// CSSoundChannelMCI //

CSSoundChannelMCI::CSSoundChannelMCI(uint uiInstIdx, uint uiDataId, uint uiChannelId, CSoundMCI *pSoundMCI):
CInstancedObj(uiInstIdx), _uiChannelId(uiChannelId), _pSoundMCI(pSoundMCI), _pMusicMCI(NULL),
bAquired(false), bGlobalPaused(false)
{
	string	alias = "channel_" + IntToStr(uiChannelId),
			file_name = MMFiles[uiDataId].stFileName + "+ type waveaudio alias " + alias;
	
	CurrentMMFileId = uiDataId;
	if(0 == mciSendStringA(("open "+file_name).c_str(), NULL, 0, NULL))
		_pMusicMCI = new CMusicMCI(InstIdx(), mciGetDeviceIDA(alias.c_str()));
}

CSSoundChannelMCI::~CSSoundChannelMCI()
{
	if(_pMusicMCI)
		delete _pMusicMCI;
}

bool CSSoundChannelMCI::Play(bool bLooped)
{
	if(!_pMusicMCI)
		return false;

	return _pMusicMCI->Play(bLooped) == S_OK;
}

void CSSoundChannelMCI::Pause(bool bPaused)
{
	if(!_pMusicMCI)
		return;

	_pMusicMCI->Pause(bPaused);
}

bool CSSoundChannelMCI::IsPlaying()
{
	if(!_pMusicMCI)
		return false;

	bool ret;
	_pMusicMCI->IsPlaying(ret);

	return ret;
}

HRESULT CALLBACK CSSoundChannelMCI::PlayOrPause()
{
	if(!_pMusicMCI)
		return S_FALSE;

	bool playing;
	_pMusicMCI->IsPlaying(playing);

	return _pMusicMCI->Pause(playing);
}

HRESULT CALLBACK CSSoundChannelMCI::Stop()
{
	if(!_pMusicMCI)
		return S_FALSE;

	return _pMusicMCI->Stop();
}

HRESULT CALLBACK CSSoundChannelMCI::IsPlaying(bool &bIsPlaying)
{
	if(!_pMusicMCI)
		return S_FALSE;

	return _pMusicMCI->IsPlaying(bIsPlaying);
}

HRESULT CALLBACK CSSoundChannelMCI::SetVolume(uint uiVolume)
{
	if(!_pMusicMCI)
		return S_FALSE;

	return _pMusicMCI->SetVolume((float)uiVolume * _pSoundMCI->GetMasterVolume());
}

HRESULT CALLBACK CSSoundChannelMCI::GetVolume(uint &uiVolume)
{
	if(!_pMusicMCI)
		return S_FALSE;

	uint volume;

	HRESULT ret = _pMusicMCI->GetVolume(volume);

	uiVolume = (float)volume / _pSoundMCI->GetMasterVolume();

	return ret;
}

HRESULT CALLBACK CSSoundChannelMCI::SetPan(int iPan)
{
	//Unsupported feature
	return S_FALSE;
}

HRESULT CALLBACK CSSoundChannelMCI::GetPan(int &iPan)
{
	//Unsupported feature
	iPan = 0;
	return S_FALSE;
}

HRESULT CALLBACK CSSoundChannelMCI::SetFrequency(uint32 uiFreq)
{
	//Unsupported feature
	return S_FALSE;
}

HRESULT CALLBACK CSSoundChannelMCI::GetFrequency(uint32 &uiFreq)
{
	//Unsupported feature
	uiFreq = 0;
	return S_FALSE;
}

HRESULT CALLBACK CSSoundChannelMCI::SetPosition(const TPoint3 &stCoords)
{
	//Unsupported feature
	return S_FALSE;
}

HRESULT CALLBACK CSSoundChannelMCI::GetPosition(TPoint3 &stCoords)
{
	//Unsupported feature
	stCoords = TPoint3();
	return S_FALSE;
}

HRESULT CALLBACK CSSoundChannelMCI::SetEffects(E_SOUND_CHANNEL_EFFECTS eFlags)
{
	//Unsupported feature
	return S_FALSE;
}

HRESULT CALLBACK CSSoundChannelMCI::Unaquire()
{
	bAquired = false;
	return S_OK;
}

// CSSampleMCI //

CSSampleMCI::CSSampleMCI(uint uiInstIdx, uint uiDataId, CSoundMCI *pSoundMCI):
CInstancedObj(uiInstIdx), _uiDataId(uiDataId), _uiChannelId(-1), _pSoundMCI(pSoundMCI)
{}

CSSampleMCI::~CSSampleMCI()
{
	MMFiles[_uiDataId] = TMMFile();
}

HRESULT CALLBACK CSSampleMCI::Play()
{
	if(!_pSoundMCI->CreateChannel(_uiDataId, _uiChannelId))
		return E_ABORT;
	else
		return _pSoundMCI->GetChannel(_uiChannelId)->Play(false) ? S_OK : E_ABORT;
}

HRESULT CALLBACK CSSampleMCI::PlayEx(ISoundChannel *&pSndChnl, E_SOUND_SAMPLE_PARAMS eFlags)
{
	if(!_pSoundMCI->CreateChannel(_uiDataId, _uiChannelId))
		return E_ABORT;
	else
	{
		CSSoundChannelMCI *p_channel = _pSoundMCI->GetChannel(_uiChannelId);
		pSndChnl = (ISoundChannel*)p_channel;
		p_channel->bAquired = true;
		return p_channel->Play(eFlags & SSP_LOOP) ? S_OK : E_ABORT;
	}
}

HRESULT CALLBACK CSSampleMCI::Free()
{
	bool can_delete;

	Core()->pResMan()->RemoveResource(this, can_delete);

	if(can_delete)
	{
		delete this;
		return S_OK;
	}
	else
		return S_FALSE;
}

HRESULT CALLBACK CSSampleMCI::GetType(E_ENG_OBJ_TYPE &eObjType)
{
	eObjType = EOT_SOUND_SAMPLE;
	return S_OK;
}

// CSoundMCI //

CSoundMCI::CSoundMCI(uint uiInstIdx):
CInstancedObj(uiInstIdx), _uiMaxChannels(32), _fMasterVolume(1.f)
{
	Core()->pResMan()->RegisterFileFormat("wav", EOT_SOUND_SAMPLE, "WAVe (PCM) uncompressed sound files.", &_s_LoadSoundWAV, (void*)this);
	mmioInstallIOProc(mmioFOURCC('W', 'A', 'V', ' '), (LPMMIOPROC)_s_IOProc, MMIO_INSTALLPROC);
	LOG("Sound Subsystem initialized.",LT_INFO);
}

CSoundMCI::~CSoundMCI()
{
	for(size_t i = 0; i < _channels.size(); i++)
		delete _channels[i];

	mmioInstallIOProc(mmioFOURCC('W', 'A', 'V', ' '), NULL, MMIO_REMOVEPROC);
	LOG("Sound Subsystem finalized.",LT_INFO);
}

LRESULT CALLBACK CSoundMCI::_s_IOProc(LPMMIOINFO lpMMIOInfo, UINT uMessage, LPARAM lParam1, LPARAM lParam2)
{
	switch(uMessage)
	{
	case MMIOM_OPEN:
		lpMMIOInfo->adwInfo[0] = CurrentMMFileId;
		lpMMIOInfo->lDiskOffset = 0;
		return 0;

	case MMIOM_CLOSE:
		return 0;

	case MMIOM_READ:		 
		if(!MMFiles[lpMMIOInfo->adwInfo[0]].pData)
			return -1;
		memcpy((void *)lParam1,  MMFiles[lpMMIOInfo->adwInfo[0]].pData + lpMMIOInfo->lDiskOffset, lParam2);
		lpMMIOInfo->lDiskOffset += lParam2;
		return lParam2;

	case MMIOM_SEEK:
		switch(lParam2)
		{
		case SEEK_SET:
			lpMMIOInfo->lDiskOffset = lParam1;
			break;

		case SEEK_CUR:
			lpMMIOInfo->lDiskOffset += lParam1;
			if(lpMMIOInfo->lDiskOffset > MMFiles[lpMMIOInfo->adwInfo[0]].uiFileSize)
				return -1;
			break;

		case SEEK_END:
			lpMMIOInfo->lDiskOffset = MMFiles[lpMMIOInfo->adwInfo[0]].uiFileSize - 1 - lParam1;
			break;
		}
		return lpMMIOInfo->lDiskOffset;

	default:
		return -1;
	}
}

bool CSoundMCI::CreateChannel(uint uiDataId, uint &uiChannelId)
{
	uiChannelId = -1;

	if(uiDataId >= MMFiles.size())
		return false;

	int idx = -1;
	for(size_t i = 0; i < _channels.size(); i++)
		if(!_channels[i]->bAquired && !_channels[i]->bGlobalPaused && !_channels[i]->IsPlaying())
		{
			delete _channels[i];
			idx = i;
			break;
		}

	if(idx == -1)
	{
		idx = _channels.size();

		if(_uiMaxChannels > idx + 1)
			_channels.push_back(new CSSoundChannelMCI(InstIdx(), uiDataId, idx, this));
		else
			return false;
	}
	else
		_channels[idx] = new CSSoundChannelMCI(InstIdx(), uiDataId, idx, this);

	uiChannelId = idx;

	return true;
}

CSSoundChannelMCI* CSoundMCI::GetChannel(uint uiChannelId)
{
	if(uiChannelId > _channels.size())
		return NULL;

	return _channels[uiChannelId];
}

bool CSoundMCI::_LoadSoundWAV(IFile *pFile, ISoundSample *&rpSSample)
{
	MMFiles.push_back(TMMFile(pFile));
	rpSSample = (ISoundSample*)(new CSSampleMCI(InstIdx(), MMFiles.size() - 1,  this));
	return true;
}

HRESULT CALLBACK CSoundMCI::SetMasterVolume(uint uiVolume)
{
	if(uiVolume > 100 || uiVolume < 1)
		return E_INVALIDARG;

	float prev_volume = _fMasterVolume;
	_fMasterVolume = (float)uiVolume/100.f;

	for(size_t i = 0; i < _channels.size(); i++)
	{
		uint volume;
		_channels[i]->GetVolume(volume);
		_channels[i]->SetVolume(volume);
	}

	return S_OK;
}

HRESULT CALLBACK CSoundMCI::SetMaxChannelsCount(uint uiCount)
{
	if(uiCount > 256)
		return E_INVALIDARG;
	else
		_uiMaxChannels = uiCount;

	return S_OK;
}

HRESULT CALLBACK CSoundMCI::PauseAllChannels(bool bPaused)
{
	for(size_t i = 0; i < _channels.size(); i++)
	{
		_channels[i]->bGlobalPaused = bPaused;
		_channels[i]->Pause(bPaused);
	}

	return S_OK;
}

HRESULT CALLBACK CSoundMCI::StopAllChannels()
{
	for(size_t i = 0; i < _channels.size(); i++)
		_channels[i]->Stop();

	return S_OK;
}

HRESULT CALLBACK CSoundMCI::SetListnerPosition(const TPoint3 &stCoords)
{
	//Unsupported feature
	return S_FALSE;
}

HRESULT CALLBACK CSoundMCI::GetListnerPosition(TPoint3 &stCoords)
{
	//Unsupported feature
	stCoords = TPoint3();
	return S_FALSE;
}

HRESULT CALLBACK CSoundMCI::SetListnerOrientation(const TPoint3 &stDir, const TPoint3 &stUp)
{
	//Unsupported feature
	return S_FALSE;
}

HRESULT CALLBACK CSoundMCI::GetListnerOrientation(TPoint3 &stDir, TPoint3 &stUp)
{
	//Unsupported feature
	stDir = TPoint3();
	stUp = TPoint3();
	return S_FALSE;
}

HRESULT CALLBACK CSoundMCI::GetType(E_ENGINE_SUB_SYSTEM &eSubSysType)
{
	eSubSysType = ESS_SOUND;
	return S_OK;
}

bool CALLBACK CSoundMCI::_s_LoadSoundWAV(IFile *pFile, IEngBaseObj *&prObj, uint uiLoadFlags, void *pParametr)
{
	ISoundSample *pss = NULL;
	bool ret = PTHIS(CSoundMCI)->_LoadSoundWAV(pFile, pss);
	if(ret)
		prObj = (IEngBaseObj *&)pss;
	return ret;
}