/**
\author		Korotkov Andrey aka DRON
\date		15.03.2011 (c)Korotkov Andrey

This file is a part of DGLE2 project and is distributed
under the terms of the GNU Lesser General Public License.
See "DGLE2.h" for more details.
*/

#if _MSC_VER > 1000
#pragma once
#endif

#ifndef _SOUNDMCI_H
#define _SOUNDMCI_H

#include "..\stdafx.h"
#include "MusicMCI.h" //also includes MMSystem.h and Winmm.lib
#include "Core.h"

class CCore;
class CMusicMCI;
class CSoundMCI;

class CSSoundChannelMCI : public CInstancedObj, public ISoundChannel
{
	uint		_uiChannelId;
	CMusicMCI	*_pMusicMCI;
	CSoundMCI	*_pSoundMCI;

public:
	bool bAquired, bGlobalPaused;

	CSSoundChannelMCI(uint uiInstIdx, uint uiDataId, uint uiChannelId, CSoundMCI *pSoundMCI);
	~CSSoundChannelMCI();

	bool Play(bool bLooped);
	void Pause(bool bPaused);
	bool IsPlaying();

	HRESULT CALLBACK PlayOrPause();
	HRESULT CALLBACK Stop();
	HRESULT CALLBACK IsPlaying(bool &bIsPlaying);
	HRESULT CALLBACK SetVolume(uint uiVolume);
	HRESULT CALLBACK GetVolume(uint &uiVolume);
	HRESULT CALLBACK SetPan(int iPan);
	HRESULT CALLBACK GetPan(int &iPan);
	HRESULT CALLBACK SetFrequency(uint32 uiFreq);
	HRESULT CALLBACK GetFrequency(uint32 &uiFreq);
	HRESULT CALLBACK SetPosition(const TPoint3 &stCoords);
	HRESULT CALLBACK GetPosition(TPoint3 &stCoords);
	HRESULT CALLBACK SetEffects(E_SOUND_CHANNEL_EFFECTS eFlags);
	HRESULT CALLBACK Unaquire();

	IDGLE2_BASE_IMPLEMENTATION(ISoundChannel)
};

class CSSampleMCI: public CInstancedObj, public ISoundSample
{
	CSoundMCI *_pSoundMCI;
	uint _uiDataId, _uiChannelId;

public:

	CSSampleMCI(uint uiInstIdx, uint uiDataId, CSoundMCI *pSoundMCI);
	~CSSampleMCI();

	HRESULT CALLBACK Play();
	HRESULT CALLBACK PlayEx(ISoundChannel *&pSndChnl, E_SOUND_SAMPLE_PARAMS eFlags);

	HRESULT CALLBACK Free();
	HRESULT CALLBACK GetType(E_ENG_OBJ_TYPE &eObjType);

	IDGLE2_BASE_IMPLEMENTATION1(ISoundSample, IEngBaseObj)
};

class CSoundMCI: public CInstancedObj, public ISound
{
	bool					_LoadSoundWAV(IFile *pFile, ISoundSample *&rpSSample);
	static bool CALLBACK	_s_LoadSoundWAV(IFile *pFile, IEngBaseObj *&prObj, uint uiLoadFlags, void *pParametr);
	static LRESULT CALLBACK	_s_IOProc(LPMMIOINFO lpMMIOInfo, UINT uMessage, LPARAM lParam1, LPARAM lParam2);

	uint	_uiMaxChannels;
	float	_fMasterVolume;

	vector<CSSoundChannelMCI *> _channels;

public:

	CSoundMCI(uint uiInstIdx);
	~CSoundMCI();

	bool CreateChannel(uint uiDataId, uint &uiChannelId);
	float GetMasterVolume(){ return _fMasterVolume; }
	CSSoundChannelMCI* GetChannel(uint uiChannelId);

	HRESULT CALLBACK SetMasterVolume(uint uiVolume);
	HRESULT CALLBACK SetMaxChannelsCount(uint uiCount);
	HRESULT CALLBACK PauseAllChannels(bool bPaused);
	HRESULT CALLBACK StopAllChannels();
	HRESULT CALLBACK SetListnerPosition(const TPoint3 &stCoords);
	HRESULT CALLBACK GetListnerPosition(TPoint3 &stCoords);
	HRESULT CALLBACK SetListnerOrientation(const TPoint3 &stDir, const TPoint3 &stUp);
	HRESULT CALLBACK GetListnerOrientation(TPoint3 &stDir, TPoint3 &stUp);

	HRESULT CALLBACK GetType(E_ENGINE_SUB_SYSTEM &eSubSysType);

	IDGLE2_BASE_IMPLEMENTATION(ISound)
};

#endif //_SOUNDMCI_H

